@@ -23,14 +23,10 @@ use Curses::UI::Widget;
 # to our calling this unless somebody is being really, really bad.
 BEGIN { run POE::Kernel }
 
-*VERSION = \0.02801;
+*VERSION = \0.031;
 our $VERSION;
 
 use constant TOP => -1;
-use constant PARENT => -2;
-use constant ROOT => 0;
-
-my @ModalObject;
 
 sub import {
     my $caller = caller;
@@ -41,86 +37,159 @@ sub import {
     eval "package $caller; use POE;";
 }
 
-sub new { 
-    my $RootObject = &Curses::UI::new(@_);
-    push @ModalObject, $RootObject; 
+# XXX We assume that there will never be two Curses::UI::POE sessions.
+my @modal_objects;
+my @modal_callbacks;
 
-    my ($class, %Options) = @_;
+# The session needed to make curses run in POE.
+sub new {
+    my ($type, %options) = @_;
+    my $self = &Curses::UI::new(@_);
+#   my $self = bless Curses::UI->new, $type;
+#   my $self = bless &Curses::UI::new(@_), $type;
 
-    $Options{inline_states} ||= {};
+    # I have to do this here, because if our first order of business is a
+    # dialog then the _start event will be too late.  This self reference is
+    # just so we can stack and peel onto the list of modal objects, and get to
+    # ourselves when we reach the top.
+    push @modal_objects, $self;
+
+    $self->{options}            = \%options;
+    $self->{__start_callback}   = delete $options{inline_states}{_start};
+
+    delete $options{package_states}{_start};
+
+    # Default so we don't get a warning about using undef as an array.
+    $options{object_states}   ||= [];
+    $options{inline_states}   ||= {};
+    $options{options}         ||= {};
 
     POE::Session->create
-        ( inline_states => {
-            %{ $Options{inline_states} },
+        ( options        => $options{options},
+          args           => $options{args},
+          inline_states  => $options{inline_states},
 
-            _start  => sub {
-                $_[KERNEL]->select(\*STDIN, "-keyin");
+          object_states  => [
+            @{ $options{object_states} },
+            $self, [ qw( _start keyin timer shutdown ) ]
+          ],
+          
+          # This is to maintain backward compatibility.
+          heap => $self );
 
-                # Turn blocking back on for STDIN.  Some Curses
-                # implementations don't deal well with non-blocking STDIN.
-                my $flags = fcntl STDIN, F_GETFL, 0 or die $!;
-                fcntl STDIN, F_SETFL, $flags & ~O_NONBLOCK or die $!;
+    # Copy the no-output option
+    $self->{-no_output} = $options{-no_output} || 0;
 
-                set_read_timeout($ModalObject[TOP]);
-
-                $_[HEAP] = $RootObject;
-
-                $Options{inline_states}{_start}(@_)
-                    if defined $Options{inline_states}{_start};
-            },
-
-            -keyin  => sub {
-                unless ($#ModalObject) {
-                    $RootObject->do_one_event;
-                }
-                else {
-                    $ModalObject[TOP]->root->do_one_event($ModalObject[TOP]);
-                }
-                Curses::curs_set($ModalObject[ROOT]->{-cursor_mode});
-
-#                if (my $key = $_[HEAP]->get_key(0)) {
-#                    $RootObject->feedkey($key) unless $key eq '-1';
-#                    $RootObject->do_one_event;
-#                }
-            },
-
-            -timer  => sub {
-                $ModalObject[TOP]->do_timer;
-                Curses::curs_set($ModalObject[ROOT]->{-cursor_mode});
-
-                set_read_timeout($ModalObject[TOP]);
-            },
-
-            shutdown => sub {
-                $_[KERNEL]->select(\*STDIN);
-            },
-          },
-
-          heap => $RootObject,
-        );
-
-     $RootObject->{-no_output} = $Options{-no_output} || 0;
-
-     return $RootObject;
+    return $self;
 }
 
-sub MainLoop { 
-    unless ($ModalObject[TOP]) {
-        die "MailLoop: Curses::UI::rootobject not created.";
+sub _start {
+    my ($self, $kernel) = @_[ OBJECT, KERNEL ];
+
+    $kernel->select(\*STDIN, "keyin");
+
+    # Turn blocking back on for STDIN.  Some Curses
+    # implementations don't deal well with non-blocking STDIN.
+    my $flags = fcntl STDIN, F_GETFL, 0 or die $!;
+    fcntl STDIN, F_SETFL, $flags & ~O_NONBLOCK or die $!;
+
+    # If we're in a dialog, then the TOP modal object is more appropriate than
+    # $self, although if we're not in a dialog $self is what this actually is.
+    set_read_timeout($modal_objects[TOP]);
+
+    # Unmask...
+    $self->{__start_callback}(@_)
+        if defined $self->{__start_callback};
+}
+
+sub _clear_modal_callback {
+    my ($self) = @_;
+
+    my $top     = pop @modal_objects;
+
+    # Reset focus
+    $top->{-focus} = 0;
+
+    # Dispatch callback.
+    my $args    = pop @modal_callbacks;
+    my $sub     = shift @$args;
+    &{$sub}(@$args);
+}
+
+sub keyin {
+    my ($self, $kernel) = @_[ OBJECT, KERNEL ];
+
+    unless ($#modal_objects) {
+        $self->do_one_event;
+    }
+    else {
+        # dispatch the event to the top-most modal object, or the root.
+        $self->do_one_event($modal_objects[TOP]);
+
+# I didn't originally do this here, I'm not quite sure what I'm up to...
+#
+#   # If this is a callback modal focus widget, and we lost modal focus,
+#   # execute the callback an clear the level in the stack.
+#       $self->_clear_modal_callback 
+#           unless $modal_objects[TOP]->{-has_modal_focus};
+
+# This other wierdness seems unnecessary.
+#       $top_object->root->do_one_event($top_object);
     }
 
-    $ModalObject[TOP]->mainloop 
+# This was a hack to make sure to pick up the extra events when things got out
+# of sync.  I'm not sure if I need it.  But let's try getting C::U::P working
+# first.
+#   if (my $key = $self->get_key(0)) {
+#       $self->feedkey($key) unless $key eq "-1";
+#       $self->do_one_event;
+#   }
+
+    # Set the root cursor mode
+    unless ($self->{-no_output}) {
+        Curses::curs_set($self->{-cursor_mode});
+    }
+}
+
+sub timer {
+    my ($self) = @_;
+
+    # dispatch the event to the top-most modal object, or the root.
+    my $top_object = $modal_objects[TOP];
+
+    $top_object->do_timer;
+
+    # Set the root cursor mode.
+    unless ($self->{-no_output}) {
+        Curses::curs_set($self->{-cursor_mode});
+    }
+
+    set_read_timeout($top_object);
+
+# Looks like older versions didn't support callbackmodalfocus, whatever that
+# is.
+# I'm not sure what the deal is with the callbackmodalfocus shit...
+#   $self->_clear_modal_callback unless $top_object->{-has_modal_focus};
+}
+
+sub shutdown {
+    my ($kernel) = $_[ KERNEL ];
+
+    # Unselect stdin
+    $kernel->select(\*STDIN);
 }
 
 sub mainloop {
-    my $this = shift;
+    my ($this) = @_;
 
-    $this->focus(undef, 1);
-    $this->draw;
+    unless ($this->{-no_output}) {
+        $this->focus(undef, 1);
+        $this->draw;
 
-    Curses::doupdate;
+        Curses::doupdate;
+    }
 
-    run POE::Kernel;
+    POE::Kernel->run;
 }
 
 sub set_read_timeout {
@@ -136,7 +205,7 @@ sub set_read_timeout {
             $new_timeout < $config->{-time};
     }
 
-    $poe_kernel->delay(-timer => $new_timeout) if $new_timeout >= 0;
+    $poe_kernel->delay(timer => $new_timeout) if $new_timeout >= 0;
 
     # Force the read timeout to be 0, so Curses::UI polls.
     $this->{-read_timeout} = 0;
@@ -144,18 +213,48 @@ sub set_read_timeout {
     return $this;
 }
 
-# Redefine the modalfocus loop because it sucks.
 {
-    no warnings "redefine"; 
-    sub Curses::UI::Widget::modalfocus () {
-        my $this = shift;
+    no warnings "redefine";
+    # None of this work's if POE isn't running...
+    # Redefine the callbackmodalfocus to ensure that callbacks and objects make
+    # it on to our own private stack.
+    sub Curses::UI::Widget::callbackmodalfocus {
+        my ($this, $cb) = @_;
 
         # "Fake" focus for this object.
         $this->{-has_modal_focus} = 1;
         $this->focus;
         $this->draw;
 
-        push @ModalObject, $this;
+        push @modal_objects, $this;
+
+        if (defined $cb) {
+            # They need a callback, so register it.
+            push @modal_callbacks, $cb;
+        } else {
+            # Push a null callback.
+            push @modal_callbacks, [sub { }];
+        }
+
+        # We assume our callers are going to return immediately back to the
+        # main event loop, so we don't need a recursive call.       
+        return;
+    }
+
+    # Let modalfocus() be a reentrant into the POE Kernel.  This is stackable,
+    # so it should not impact other behaviors, and POE keeps chugging along
+    # uneffected.  This is a modal focus without a callback, this method does
+    # not return until the modal widget get's cleared out.
+    sub Curses::UI::Widget::modalfocus () {
+        my ($this) = @_;
+
+        # "Fake" focus for this object.
+        $this->{-has_modal_focus} = 1;
+        $this->focus;
+        $this->draw;
+
+        push @modal_objects, $this;
+        push @modal_callbacks, undef;
 
         # This is reentrant into the POE::Kernel 
         while ( $this->{-has_modal_focus} ) {
@@ -164,7 +263,8 @@ sub set_read_timeout {
 
         $this->{-focus} = 0;
 
-        pop @ModalObject;
+        pop @modal_callbacks;
+        pop @modal_objects;
 
         return $this;
     }
